#include "auth.h"
#include "config.h"

#include <hiredis/hiredis.h>
#include <uuid/uuid.h>
#include <string.h>
#include <errno.h>
#include <tomcrypt.h>

#define UUID_SIZE 37
#define SHA512_DIGEST_SIZE 64

#define FAIL_AUTH() do { \
    strcpy(session_id, ""); \
    retval = 0; \
    goto defer; \
} while (0)

/* redis auth defaults */
static struct timeval ra_connection_timeout = {1, 500000}; /* 1.5 secs */ 

static char *ra_hostname = "ra-socket";
static char *ra_magic = "$z$";


int
load_auth_config()
{
    CONF_ROOT_LOOKUP(string, ra_hostname);
    CONF_ROOT_LOOKUP(string, ra_magic);

    CONF_ROOT_LOOKUP(int, user_max_length);
    CONF_ROOT_LOOKUP(int, password_max_length);

    CONF_ROOT_LOOKUP(int, session_expire);
    
    return 0;
}

int
user_by_session_id(const char *session_id, char *username)
{
    redisContext *ctx;
    redisReply *reply;
    uuid_t uu_tmp;
    int retval = 0;

    ctx = NULL;
    reply = NULL;

    if (uuid_parse(session_id, uu_tmp) < 0) {
        retval = -EINVAL;
        goto defer;
    }

    ctx = redisConnectUnixWithTimeout(ra_hostname, ra_connection_timeout);

    if (!ctx || ctx->err) {
        retval = -EIO;
        goto defer;
    }

    reply = redisCommand(ctx, "GET username:%s", session_id);
    if (reply->type != REDIS_REPLY_STRING) {
        strcpy(username, "");
    } else {
        if (reply->len > user_max_length + 1) {
            retval = -EINVAL;
            goto defer;
        }
        strcpy(username, reply->str);
    }

defer:
    redisFree(ctx);
    freeReplyObject(reply);
    return retval;
}

int
authenticate(const char *username, const char *password, char *session_id)
{
    redisContext *ctx;
    redisReply *salt_reply, *hash_reply, *sid_reply;
    char *sid;
    unsigned char db_hash[SHA512_DIGEST_SIZE], got_hash[SHA512_DIGEST_SIZE];
    uuid_t uu_tmp;
    hash_state md;
    int retval = 0;

    ctx = NULL;
    salt_reply = hash_reply = sid_reply = NULL;

    if (strlen(username) > user_max_length || strlen(password) > password_max_length) {
        retval = -EINVAL;
        goto defer;
    }

    ctx = redisConnectUnixWithTimeout(ra_hostname, ra_connection_timeout);

    if (!ctx || ctx->err) {
        retval = -EIO;
        goto defer;
    }

    salt_reply = redisCommand(ctx, "GET salt:%s", username);

    if (salt_reply->type != REDIS_REPLY_STRING) {
        FAIL_AUTH();
    }

    hash_reply = redisCommand(ctx, "GET hash:%s", username);

    if (hash_reply->type != REDIS_REPLY_STRING || hash_reply->len != SHA512_DIGEST_SIZE) {
        FAIL_AUTH();
    }

    memcpy(db_hash, hash_reply->str, SHA512_DIGEST_SIZE);

    sha512_init(&md);
    
    sha512_process(&md, ra_magic, strlen(ra_magic));
    sha512_process(&md, password, strlen(password));
    sha512_process(&md, salt_reply->str, strlen(salt_reply->str));

    sha512_done(&md, got_hash);

    if (memcmp(db_hash, got_hash, SHA512_DIGEST_SIZE)) {
        FAIL_AUTH();
    } else {
        sid_reply = redisCommand(ctx, "GET session:%s", username);
        if (sid_reply->type != REDIS_REPLY_STRING) {
            /* assuming there is no session for this user */
            freeReplyObject(sid_reply);

            uuid_generate(uu_tmp);
            uuid_unparse(uu_tmp, session_id);

            sid_reply = redisCommand(ctx, "SET session:%s %s", username, session_id);
            freeReplyObject(sid_reply);
            sid_reply = redisCommand(ctx, "SET username:%s %s", session_id, username);
            freeReplyObject(sid_reply);
        } else {
            strcpy(session_id, sid_reply->str);
        }
        sid_reply = redisCommand(ctx, "EXPIRE session:%s %d", username, session_expire);
        freeReplyObject(sid_reply);
        sid_reply = redisCommand(ctx, "EXPIRE username:%s %d", session_id, session_expire);
        retval = 1;
    }

defer:
    redisFree(ctx);
    freeReplyObject(salt_reply);
    freeReplyObject(hash_reply);
    freeReplyObject(sid_reply);
    return retval;
}
