#include "config.h"
#include "string_buffer.h"
#include "utils.h"
#include "comment.h"
#include "cgic.h"
#include "driver.h"
#include "drivers/unix_fs/unix_fs_driver.h"
#include "auth.h"

#include <stdlib.h>
#include <time.h>
#include <stdio.h>
#include <string.h>
#include <syslog.h>

#define UUID_SIZE 37

char session_id[UUID_SIZE], *authorized_user_sid;

int
page_by_id(int id)
{
    return id / comments_per_page + 1;
}

void
render_comment(const Comment *comment)
{
    char *retval;
    int rid;
    
    fprintf(cgiOut, "<div class=\"comment\" id=\"comment_%d\">\n", comment->id);
    
    /* begin header */
    fprintf(cgiOut, "<div class=\"comment-header\">\n");
    fprintf(cgiOut, "Posted by: %s\n", comment->header->user_displayname);

    rid = comment->header->reply_id;
    if (rid > 0) {
        fprintf(cgiOut, "<a href=\"?page=%d#comment_%d\"> in reply to </a>", page_by_id(rid), rid);
    }

    fprintf(cgiOut, "</div>\n");
    /* end header */

    fprintf(cgiOut, "%s", comment->text);

    fprintf(cgiOut, "</div>\n");
}

void
allocate_header(CommentHeader *header)
{
    header->user_sid = malloc(max_name_size);
    header->user_displayname = malloc(max_name_size);
}

void
free_header(CommentHeader *header)
{
    free(header->user_sid);
    free(header->user_displayname);
}

void
allocate_comment(Comment *comment)
{
    comment->header = malloc(sizeof(CommentHeader));
    allocate_header(comment->header);
}

void
free_comment(Comment *comment)
{
    free_header(comment->header);
    free(comment->header);
    free(comment->text);
}

int
fetch_comment(int id, Driver *driver, void *driver_data, Comment *comment)
{
    int retval;
    comment->id = id;

    allocate_comment(comment);

    retval = driver->get_header(driver_data, comment->header, id);
    if (retval < 0) {
        goto defer;
    }
    comment->text = malloc(comment->header->text_length + 1);

    retval = driver->get_text(driver_data, comment->text, id);
defer:
    if (retval < 0) {
        free_comment(comment);
    }
    return retval;
}

void
print_submit_form()
{
    fputs("<form action=\"\" method=\"POST\">\n"
        "<label>Your name:</label>\n"
        "<input type=\"text\" name=\"displayname\"><br>\n"
        "<label>Leave comment: </label><br>\n"
        "<input type=\"text\" name=\"text\"><br>\n"
        "<input type=\"submit\" name=\"submit\" value=\"Submit\">\n"
    "</form>\n", cgiOut);
}

/* each page stores comments with ids in [P * (page - 1) + 1, P * page + 1)
 * if page is not specified, then we'll just print last P comments (for now)
 * TODO: print errors in div
 */
void
print_page()
{
    int page, max_id;
    int id_begin, id_end, i;
    Comment comment;

    cgiFormInteger("page", &page, 0);
    if ((max_id = unix_fs_driver_get_max_id(&driver_data)) < 0) {
        fprintf(cgiOut, "Error fetching comments\n");
        return;
    }

    if (page <= 0) {
        id_end = max_id;
        id_begin = (max_id > comments_per_page ? max_id - comments_per_page : 1);
    } else {
        /* max_id > comments_per_page * (page - 1) + 1
         * max_id - 1 > comments_per_page * (page - 1)
         * perform this check without overflow
         */
        if ((max_id + comments_per_page - 1) / comments_per_page <= (page - 1)) {
            fprintf(cgiOut, "There is no such page\n");
            return;
        }
        id_begin = comments_per_page * (page - 1) + 1;
        id_end = id_begin + comments_per_page;
        if (id_end > max_id) {
            id_end = max_id;
        }
    }

    for (i = id_begin; i != id_end; ++i) {
        if (fetch_comment(i, &driver, &driver_data, &comment) < 0) {
            fprintf(cgiOut, "Failed to fetch comment %d\n", i);
        } else {
            render_comment(&comment);
            free_comment(&comment);
        }
    }
}

/* TODO: restore fields for user no to lose data */
int
handle_submitted_comment()
{
    char displayname[max_name_size], text[max_comment_size], *sanitized_text;
    int rid, retval;
    cgiFormResultType err;
    CommentHeader header;

    err = cgiFormString("text", text, max_comment_size);
    if (err == cgiFormTruncated) {
        fprintf(cgiOut, "Comment too long (max %d bytes)\n", max_comment_size);
        goto defer;
    }
    if (err == cgiFormNotFound) {
        fprintf(cgiOut, "Comment text not provided\n");
        goto defer;
    }

   err = cgiFormString("displayname", displayname, max_name_size);
   if (err == cgiFormTruncated) {
       fprintf(cgiOut, "Name too long (max %d bytes)\n", max_name_size);
       goto defer;
   }
   if (err == cgiFormNotFound) {
       fprintf(cgiOut, "Name not provided\n");
       goto defer;
   }

   if (contain_special(displayname)) {
       err = cgiFormTruncated;
       fprintf(cgiOut, "Name must not contain HTML special characters\n");
       goto defer;
   }

   cgiFormInteger("reply-to", &rid, 0);

   sanitized_text = mk_specialchars(text);

   time(&header.creation_time);
   header.reply_id = rid;
   header.text_length = strlen(sanitized_text);
   if (strcmp(authorized_user_sid, "")) {
       header.user_sid = authorized_user_sid;
   } else {
       header.user_sid = "web/anonymous";
   }
   header.user_displayname = displayname;

   if (driver.leave_comment(&driver_data, &header, sanitized_text) < 0) {
       fprintf(cgiOut, "Failed to leave your comment\n");
   } else {
       fprintf(cgiOut, "Comment successfully left\n");
   }

   free(sanitized_text);

defer:
   return (err == cgiFormSuccess ? 0 : -1);
}

int
cgiMain()
{
    int cfg_pld_err;
    config_init(&config);
    cfg_pld_err = preload_config((cgiArgc > 1 ? cgiArgv[1] : NULL));

    if (cfg_pld_err == CFG_PLD_INVALID) {
        syslog(LOG_ERR, "Got invalid config\n");
        syslog(LOG_ERR, "%s\n", config_error_text(&config));
        exit(EXIT_FAILURE);
    } else if (cfg_pld_err == CFG_PLD_NOT_FOUND) {
        syslog(LOG_WARNING, "Config not found, loading defaults\n");
    } else {
        load_auth_config();
    }

    CONF_ROOT_LOOKUP(int, comments_per_page);
    CONF_ROOT_LOOKUP(int, max_name_size);
    CONF_ROOT_LOOKUP(int, max_comment_size);

    CONF_ROOT_LOOKUP(int, user_max_length);
    CONF_ROOT_LOOKUP(int, password_max_length);
    
    CONF_ROOT_LOOKUP(string, qments_path);
    CONF_ROOT_LOOKUP(string, session_cookie_name);
    CONF_ROOT_LOOKUP(string, hostname);
    
    driver_data.path = qments_path;

    authorized_user_sid = malloc(user_max_length + 1);
    if (cgiCookieString(session_cookie_name, session_id, UUID_SIZE) == cgiFormSuccess) {
        user_by_session_id(session_id, authorized_user_sid);
    } else {
        authorized_user_sid[0] = '\0';
    }
    cgiHeaderContentType("text/html; charset=utf-8");

    fprintf(cgiOut, "<html>\n");

    fprintf(cgiOut, "<head>\n");
    fprintf(cgiOut, "<title> Simple discuss powered by qments </title>\n");
    fprintf(cgiOut, "</head>\n");

    fprintf(cgiOut, "<body>\n");

    if (cgiFormSubmitClicked("submit") == cgiFormSuccess && 
            !strcmp(cgiRequestMethod, "POST")) {
        handle_submitted_comment();
    }
    
    print_submit_form();
    
    fprintf(cgiOut, "<div class=\"comment-section\">\n");
    print_page();
    fprintf(cgiOut, "</div>\n");

    fprintf(cgiOut, "</body>\n");

    fprintf(cgiOut, "</html>\n");
    
    free(authorized_user_sid);
    config_destroy(&config);

    return 0;
}
